#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from tirmite import __version__
import os
import sys
import glob
import shutil
import tirmite
import argparse


def log(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def mainArgs():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Map TIR-pHMM models to genomic sequences for annotation \
        of MITES and complete DNA-Transposons.",
        prog="tirmite",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    # Input
    parser.add_argument(
        "--genome",
        type=str,
        required=True,
        help="Path to target genome that will be queried with HMMs.",
    )
    parser.add_argument(
        "--hmmDir",
        type=str,
        default=None,
        help="Directory containing pre-prepared TIR-pHMMs.",
    )
    parser.add_argument(
        "--hmmFile",
        type=str,
        default=None,
        help='Path to single TIR-pHMM file. Incompatible with "--hmmDir".',
    )
    parser.add_argument(
        "--alnDir",
        type=str,
        default=None,
        help="Path to directory containing only TIR alignments to be converted to HMM.",
    )
    parser.add_argument(
        "--alnFile",
        type=str,
        default=None,
        help='Provide a single TIR alignment to be converted to HMM. Incompatible with "--alnDir".',
    )
    parser.add_argument(
        "--alnFormat",
        default="fasta",
        choices=["clustal", "fasta", "nexus", "phylip", "stockholm"],
        help='Alignments provided with "--alnDir" or "--alnFile" are all in this format.',
    )
    parser.add_argument(
        "--pairbed",
        type=str,
        default=None,
        help="If set TIRmite will preform pairing on TIRs from custom bedfile only.",
    )
    # Alternative search method
    # parser.add_argument('--useBowtie2',action='store_true',default=False,help='If set, map short TIR to genome with bowtie2. Potentially useful for very short though highly conserved TIRs where TIR-pHMM hits return high e-values.')
    # parser.add_argument('--btTIR',type=str,default=None,help='Fasta file containing a single TIR to be mapped with bowtie2.')
    # parser.add_argument('--bowtie2',type=str,default='bowtie2',help='Set location of bowtie2 if not in PATH.')
    # parser.add_argument('--bt2build',type=str,default='bowtie2-build',help='Set location of bowtie2-build if not in PATH.')
    # parser.add_argument('--samtools',type=str,default='samtools',help='Set location of samtools if not in PATH.')
    # parser.add_argument('--bedtools',type=str,default='bedtools',help='Set location of bedtools if not in PATH.')
    # Pairing heuristics
    parser.add_argument(
        "--stableReps",
        type=int,
        default=0,
        help="Number of times to iterate pairing procedure when no additional pairs are found AND remaining unpaired hits > 0.",
    )
    # Output and housekeeping
    parser.add_argument(
        "--outdir",
        type=str,
        default=None,
        help="All output files will be written to this directory.",
    )
    parser.add_argument(
        "--prefix",
        type=str,
        default=None,
        help="Add prefix to all TIRs and Paired elements detected in this run. Useful when running same TIR-pHMM against many genomes.(Default = None)",
    )
    parser.add_argument(
        "--nopairing",
        action="store_true",
        default=False,
        help="If set, only report TIR-pHMM hits. Do not attempt pairing.",
    )
    parser.add_argument(
        "--gffOut",
        action="store_true",
        default=False,
        help="If set report features as prefix.gff3. File saved to outdir. Default: False",
    )
    parser.add_argument(
        "--reportTIR",
        default="all",
        choices=[None, "all", "paired", "unpaired"],
        help="Options for reporting TIRs in GFF annotation file.",
    )
    parser.add_argument(
        "--padlen",
        type=int,
        default=None,
        help="Extract x bases either side of TIR when writing TIRs to fasta.",
    )
    parser.add_argument(
        "--keeptemp",
        action="store_true",
        default=False,
        help="If set do not delete temp file directory.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Set syscall reporting to verbose.",
    )
    # HMMER options
    parser.add_argument(
        "--cores",
        type=int,
        default=1,
        help="Set number of cores available to hmmer software.",
    )
    parser.add_argument(
        "--maxeval",
        type=float,
        default=0.001,
        help="Maximum e-value allowed for valid hit. Default = 0.001",
    )
    parser.add_argument(
        "--maxdist",
        type=int,
        default=None,
        help="Maximum distance allowed between TIR candidates to consider valid pairing.",
    )
    parser.add_argument(
        "--nobias",
        action="store_true",
        default=False,
        help="Turn OFF bias correction of scores in nhmmer.",
    )
    parser.add_argument(
        "--matrix",
        type=str,
        default=None,
        help="Use custom DNA substitution matrix with nhmmer.",
    )
    parser.add_argument(
        "--mincov",
        type=float,
        default=0.5,
        help="Minimum valid hit length as prop of model length. Defaults to 0.5",
    )
    # Non-standard HMMER paths
    parser.add_argument(
        "--hmmpress",
        type=str,
        default="hmmpress",
        help="Set location of hmmpress if not in PATH.",
    )
    parser.add_argument(
        "--nhmmer",
        type=str,
        default="nhmmer",
        help="Set location of nhmmer if not in PATH.",
    )
    parser.add_argument(
        "--hmmbuild",
        type=str,
        default="hmmbuild",
        help="Set location of hmmbuild if not in PATH.",
    )
    args = parser.parse_args()
    return args


def missing_tool(tool_name):
    path = shutil.which(tool_name)
    if path is None:
        return [tool_name]
    else:
        return []


def main():
    """Do the work."""
    # Get cmd line args
    args = mainArgs()

    # Check for required programs.
    # tools = [args.hmmpress,args.nhmmer,args.hmmbuild,args.bowtie2,args.bt2build,args.samtools,args.bedtools]
    tools = [args.hmmpress, args.nhmmer, args.hmmbuild]

    missing_tools = []
    for tool in tools:
        missing_tools += missing_tool(tool)
    if missing_tools:
        log(
            "WARNING: Some tools required by tirmite could not be found: "
            + ", ".join(missing_tools)
        )
        log("You may need to install them to use all features.")

    # Create output and temp paths as required
    outDir, tempDir = tirmite.dochecks(args)

    # Load reference genome
    log("Log: Loading genome from: %s " % args.genome)
    genome = tirmite.importFasta(args.genome)

    # if args.useBowtie2:
    #    # Check that input fasta exists
    #    tirmite.isfile(args.btTIR)
    #    btTIRname = tirmite.getbtName(args.btTIR)
    #    # Compose bowtie map and filter commands
    #    cmds = list()
    #    cmds.append(tirmite._bowtie2build_cmd(bt2Path=args.bt2build,genome=args.genome))
    #    cmds.append(tirmite._bowtie2_cmd(bt2Path=args.bowtie2,tirFasta=args.btTIR,cores=args.cores))
    #    bam2bed_cmds,mappedPath = tirmite._bam2bed_cmd(samPath=args.samtools,bedPath=args.bedtools,tempDir=tempDir)
    #    cmds += bam2bed_cmds
    #    # Run mapp and filter
    #    tirmite.run_cmd(cmds,verbose=args.verbose,keeptemp=args.keeptemp)
    #    # Import mapping locations
    #    hitTable = tirmite.import_mapped(infile=mappedPath,tirName=btTIRname,prefix=args.prefix)
    # else:

    # Import custom TIR hits from BEDfile.
    if args.pairbed:
        # Die if no input file
        if not glob.glob(os.path.abspath(args.pairbed)):
            log("WARNING: BED file %s not found. Quitting." % args.pairbed)
            # Remove temp directory
            if not args.keeptemp:
                shutil.rmtree(tempDir)
            sys.exit(1)

        log("Log: Skipping HMM search. Using custom TIRs from file: %s" % args.pairbed)

        # Import hits from BED file
        # Format: Chrm, start, end, name, evalue, strand
        hitTable = None
        log("Log: Loading custom TIR hits from: %s" % str(args.pairbed))
        hitTable = tirmite.import_BED(
            infile=args.pairbed, hitTable=hitTable, prefix=args.prefix
        )

        # Apply hit e-value filters
        log("Log: Filtering hits with e-value > %s" % str(args.maxeval))
        hitCount = len(hitTable.index)
        hitTable = tirmite.filterHitsEval(maxeval=args.maxeval, hitTable=hitTable)
        log(
            "Log: Excluded %s hits on e-value criteria."
            % str(hitCount - len(hitTable.index))
        )
        log("Log: Remaining hits: %s " % str(len(hitTable.index)))

        # Group hits by model and chromosome (hitsDict), and initiate hit tracker hitIndex to manage pair-searching
        hitsDict, hitIndex = tirmite.table2dict(hitTable)

        # If pairing is off, just report the hits.
        if args.nopairing:
            tirmite.writeTIRs(
                outDir=outDir,
                hitTable=hitTable,
                maxeval=args.maxeval,
                genome=genome,
                padlen=args.padlen,
            )
            log("Log: Pairing is off. Reporting hits only.")
            # Remove temp directory
            if not args.keeptemp:
                shutil.rmtree(tempDir)
            sys.exit(1)

    # Else run nhmmer and load TIR hits.
    else:
        # If raw alignments provided, convert to stockholm format.
        if args.alnDir or args.alnFile:
            stockholmDir = tirmite.convertAlign(
                alnDir=args.alnDir,
                alnFile=args.alnFile,
                inFormat=args.alnFormat,
                tempDir=tempDir,
            )
        else:
            stockholmDir = None

        # If pre-built HMM provided, check correct format.
        if args.hmmFile:
            if os.path.splitext(os.path.basename(args.hmmFile))[1].lstrip(".") != "hmm":
                log("WARNING: --hmmFile has non-hmm extension. Exiting.")
                # Remove temp directory
                if not args.keeptemp:
                    shutil.rmtree(tempDir)
                sys.exit(1)

        # Compose and run HMMER commands
        cmds, resultDir, hmmDB = tirmite.cmdScript(
            hmmDir=args.hmmDir,
            hmmFile=args.hmmFile,
            alnDir=stockholmDir,
            tempDir=tempDir,
            args=args,
        )
        tirmite.run_cmd(
            cmds, verbose=args.verbose, tempDir=tempDir, keeptemp=args.keeptemp
        )

        # Die if no hits found
        if not glob.glob(os.path.join(os.path.abspath(resultDir), "*.tab")):
            log("Log: No hits found in %s . Quitting." % resultDir)
            # Remove temp directory
            if not args.keeptemp:
                shutil.rmtree(tempDir)
            sys.exit(1)

        # Import hits from nhmmer result files
        hitTable = None
        modelCount = 0
        for resultfile in glob.glob(os.path.join(os.path.abspath(resultDir), "*.tab")):
            log("Log: Loading nhmmer hits from: %s " % resultfile)
            hitTable = tirmite.import_nhmmer(
                infile=resultfile, hitTable=hitTable, prefix=args.prefix
            )
            modelCount += 1

        log(
            "Log: Imported %s hits from %s models. "
            % (str(len(hitTable.index)), str(modelCount))
        )

        # Apply hit length filters
        log("Log: Filtering hits with < %s model coverage. " % str(args.mincov))
        hitCount = len(hitTable.index)
        hitTable = tirmite.filterHitsLen(
            hmmDB=hmmDB, mincov=args.mincov, hitTable=hitTable
        )
        log(
            "Log: Excluded %s hits on coverage criteria. "
            % str(hitCount - len(hitTable.index))
        )
        log("Log: Remaining hits: %s " % str(len(hitTable.index)))

        # Apply hit e-value filters
        log("Log: Filtering hits with e-value > %s" % str(args.maxeval))
        hitCount = len(hitTable.index)
        hitTable = tirmite.filterHitsEval(maxeval=args.maxeval, hitTable=hitTable)
        log(
            "Log: Excluded %s hits on e-value criteria."
            % str(hitCount - len(hitTable.index))
        )
        log("Log: Remaining hits: %s " % str(len(hitTable.index)))

        # Group hits by model and chromosome (hitsDict), and initiate hit tracker hitIndex to manage pair-searching
        hitsDict, hitIndex = tirmite.table2dict(hitTable)

        # If pairing is off, just report the hits
        if args.nopairing:
            tirmite.writeTIRs(
                outDir=outDir,
                hitTable=hitTable,
                maxeval=args.maxeval,
                genome=genome,
                padlen=args.padlen,
            )
            log("Log: Pairing is off. Reporting hits only.")
            # Remove temp directory
            if not args.keeptemp:
                shutil.rmtree(tempDir)
            sys.exit(1)

    # Run pairing on filtered TIR set

    # Populate hitIndex with acceptible candidate partners (compatible strand and distance.)
    hitIndex = tirmite.parseHits(
        hitsDict=hitsDict, hitIndex=hitIndex, maxDist=args.maxdist
    )

    # Run iterative pairing procedure
    hitIndex, paired, unpaired = tirmite.iterateGetPairs(
        hitIndex, stableReps=args.stableReps
    )

    # Write TIR hits to fasta for each pHMM
    log("Log: Writing all valid TIR hits to fasta.")
    tirmite.writeTIRs(
        outDir=outDir,
        hitTable=hitTable,
        maxeval=args.maxeval,
        genome=genome,
        prefix=args.prefix,
        padlen=args.padlen,
    )

    # Write paired TIR hits to fasta. Pairs named as element ID + L/R tag.
    if args.reportTIR in ["all", "paired"]:
        log("Log: Writing successfully paired TIRs to fasta.")
        tirmite.writePairedTIRs(
            outDir=outDir,
            paired=paired,
            hitIndex=hitIndex,
            genome=genome,
            prefix=args.prefix,
            padlen=args.padlen,
        )

    # Extract paired hit regions (candidate TEs / MITEs) elements are stored
    # as list of gffTup objects
    pairedEles = tirmite.fetchElements(paired=paired, hitIndex=hitIndex, genome=genome)

    # Write paired-TIR features to fasta
    log("Log: Writing TIR-elements to fasta.")
    tirmite.writeElements(outDir, eleDict=pairedEles, prefix=args.prefix)

    # Write paired features to gff3, optionally also report paired/unpaired TIRs
    if args.gffOut:
        # Get unpaired TIRs
        if args.reportTIR in ["all", "unpaired"]:
            # Unpaired TIR features are stored as list of gffTup objects
            unpairedTIRs = tirmite.fetchUnpaired(hitIndex=hitIndex)
        else:
            unpairedTIRs = None

        # Set gff file path
        if args.prefix:
            gffOutPath = os.path.join(outDir, args.prefix + ".gff3")
        else:
            gffOutPath = os.path.join(outDir, "tirmite_report.gff3")
        # Write gff3
        log("Log: Writing features to gff: %s " % gffOutPath)
        tirmite.gffWrite(
            outpath=gffOutPath,
            featureList=pairedEles,
            writeTIRs=args.reportTIR,
            unpaired=unpairedTIRs,
            prefix=args.prefix,
        )

    # Remove temp directory
    if not args.keeptemp:
        shutil.rmtree(tempDir)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import tirmite as tsplit
import argparse
import shutil


def mainArgs():
    parser = argparse.ArgumentParser(
        description="Extract terminal repeats from retrotransposons (LTRs) or DNA transposons (TIRs). \
                                            Optionally, compose synthetic MITES from complete DNA transposons.",
        prog="exterminate",
    )
    parser.add_argument(
        "-i",
        "--infile",
        type=str,
        required=True,
        default=None,
        help="Multifasta containing complete elements.",
    )
    parser.add_argument(
        "-p",
        "--prefix",
        type=str,
        default=None,
        help="All output files begin with this string. \
                                        (Default: [infile name])",
    )
    parser.add_argument(
        "-d",
        "--outdir",
        type=str,
        default=None,
        help="Write output files to this directory. \
                                        (Default: cwd)",
    )
    parser.add_argument(
        "--splitmode",
        default="split",
        choices=["all", "split", "internal", "external", None],
        help='all= Report input sequence as well as internal and external segments. \
                                        split= Report internal and external segments after splitting. \
                                        internal = Report only internal segments \
                                        external = Report only terminal repeat segments. \
                                        If set to "None" then only synthetic MITES will be reported if --makemites is also set. \
                                        (Default: split)',
    )
    parser.add_argument(
        "--keeptemp",
        action="store_true",
        default=False,
        help="If set do not remove temp directory on completion.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="If set, report progress.",
    )
    parser.add_argument(
        "-m",
        "--maxdist",
        type=int,
        default=10,
        help="Terminal repeat candidates must be no more than this many bases from end of input element. (Default: 10)\
                                        Note: Increase this value if you suspect that your element is nested within some flanking sequence.",
    )
    parser.add_argument(
        "--minid",
        type=float,
        default=80.0,
        help="Minimum identity between terminal repeat pairs. As float. \
                                        (Default: 80.0)",
    )
    parser.add_argument(
        "--minterm",
        type=int,
        default=10,
        help='Minimum length for a terminal repeat to be considered. \
                                        Equivalent to nucmer "--mincluster" \
                                        (Default: 10)',
    )
    parser.add_argument(
        "--minseed",
        type=int,
        default=5,
        help='Minimum length of a maximal exact match to be included in final match cluster. \
                                        Equivalent to nucmer "--minmatch". \
                                        (Default: 5)',
    )
    parser.add_argument(
        "--diagfactor",
        type=float,
        default=0.2,
        help="Maximum diagonal difference factor for clustering of matches within nucmer, \
                                        i.e. diagonal difference / match separation (default 0.20) \
                                        Note: Increase value for greater tolerance of indels between terminal repeats.",
    )
    parser.add_argument(
        "--method",
        default="nucmer",
        choices=["blastn", "nucmer"],
        help='Select alignment method: "blastn" or "nucmer".(Default: nucmer)',
    )
    args = parser.parse_args()
    return args


def missing_tool(tool_name):
    path = shutil.which(tool_name)
    if path is None:
        return [tool_name]
    else:
        return []


def main():
    """Do the work."""
    # Get cmd line args
    args = mainArgs()

    # Check for required programs.
    tools = ["delta-filter", "nucmer", "show-coords"]
    if args.method == "blastn":
        tools.append("blastn")
    missing_tools = []
    for tool in tools:
        missing_tools += missing_tool(tool)
    if missing_tools:
        print(
            (
                "WARNING: Some tools required by Exterminate could not be found: "
                + ", ".join(missing_tools)
            )
        )
        print("You may need to install them to use all features.")

    # Create output paths as required
    outpath, tempdir = tsplit.tSplitchecks(args)

    # Load elements to be screened
    elements = tsplit.importFasta2List(args.infile)

    # If LTR mode, search for terminal repeats on same strand
    segments = tsplit.getLTRs(
        elements=elements,
        flankdist=args.maxdist,
        minterm=args.minterm,
        minseed=args.minseed,
        minid=args.minid,
        diagfactor=args.diagfactor,
        report=args.splitmode,
        temp=tempdir,
        alignTool=args.method,
        keeptemp=args.keeptemp,
        verbose=args.verbose,
    )
    tsplit.segWrite(outpath, segs=segments)

    # Remove temp directory
    if not args.keeptemp:
        shutil.rmtree(tempdir)



import os
from shlex import quote


def _bowtie2build_cmd(bt2Path="bowtie2-build", IdxPath="db/GenIdx", genome=None):
    """Construct the bowtie2-build command"""
    # Base command
    cmd = " ".join(
        ["mkdir db &&", quote(bt2Path), quote(os.path.abspath(genome)), IdxPath]
    )
    return cmd


def _bowtie2_cmd(bt2Path="bowtie2", tirFasta=None, IdxPath="db/GenIdx", cores=None):
    """Construct commands for bowtie2 mapping."""
    # bowtie2 -x genidx -f -a --very-sensitive-local -U TIR.fa --al alignments.bam
    # Base command
    cmd = " ".join(
        [
            quote(bt2Path),
            "-f -a --very-sensitive-local -x",
            IdxPath,
            "-U",
            quote(os.path.abspath(tirFasta)),
            "> alignments.sam",
        ]
    )
    # Optional set cores
    if cores:
        cmd += " --threads " + str(cores)
    return cmd


def _bam2bed_cmd(samPath="samtools", bedPath="bedtools", tempDir=None):
    """Filtering mapped reads with bedtools and samtools.
    # Fwd hits
    samtools view -b -F 0x10 alignments.sam | bedtools bamtobed -i stdin | awk -v OFS='\t' '{print $1,$2,$3,"+"}' > mapped.bed
    # Rev hits
    samtools view -b -f 0x10 alignments.sam | bedtools bamtobed -i stdin | awk -v OFS='\t' '{print $1,$2,$3,"-"}' >> mapped.bed
    """
    # Base command
    mappedPath = os.path.join(tempDir, "bowtie2mappedTIR.bed")
    cmds = list()
    # All reads not on rev strand or unmapped
    cmds.append(
        " ".join(
            [
                quote(samPath),
                "view -b -F 0x10,0x4 alignments.sam |",
                quote(bedPath),
                "bamtobed -i stdin | awk -v OFS='\\t' '{print $1,$2,$3,\"+\"}' >",
                quote(mappedPath),
            ]
        )
    )
    # Only reads on reverse strand
    cmds.append(
        " ".join(
            [
                quote(samPath),
                "view -b -f 0x10 alignments.sam |",
                quote(bedPath),
                "bamtobed -i stdin | awk -v OFS='\\t' '{print $1,$2,$3,\"-\"}' >>",
                quote(mappedPath),
            ]
        )
    )
    return cmds, mappedPath



import os
import sys
import shutil
import tempfile
import subprocess
from shlex import quote


class Error(Exception):
    pass


def decode(x):
    try:
        s = x.decode()
    except:
        return x
    return s


def _write_script(cmds, script):
    """Write commands into a bash script"""
    f = open(script, "w+")
    for cmd in cmds:
        print(cmd, file=f)
    f.close()


def syscall(cmd, verbose=False):
    """Manage error handling when making syscalls"""
    if verbose:
        print("Running command:", cmd, flush=True)
    try:
        output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as error:
        print(
            "The following command failed with exit code",
            error.returncode,
            file=sys.stderr,
        )
        print(cmd, file=sys.stderr)
        print("\nThe output was:\n", file=sys.stderr)
        print(error.output.decode(), file=sys.stderr)
        raise Error("Error running command:", cmd)
    if verbose:
        print(decode(output))


def makeBlast(seq=None, outfile=None, pid=60):
    cmd = (
        'blastn -word_size 4 -outfmt "6 qstart qend sstart send length positive pident qlen slen qframe sframe qseqid sseqid" -query '
        + quote(str(seq))
        + " -subject "
        + quote(str(seq))
        + " -out "
        + quote(str(outfile))
        + " -perc_identity "
        + str(pid)
    )
    return [cmd]


def run_blast(cmds, verbose=False):
    """Write and excute HMMER script"""
    tmpdir = tempfile.mkdtemp(prefix="tmp.", dir=os.getcwd())
    original_dir = os.getcwd()
    os.chdir(tmpdir)
    script = "run_jobs.sh"
    _write_script(cmds, script)
    syscall("bash " + script, verbose=verbose)
    os.chdir(original_dir)
    shutil.rmtree(tmpdir)


"""
Recreate pymummer-like coords file output from blast.

#Blast field, coords field, Description

qstart			[S1] 	Start of the alignment region in the reference sequence 
qend			[E1] 	End of the alignment region in the reference sequence 
sstart			[S2] 	Start of the alignment region in the query sequence 
send 			[E2] 	End of the alignment region in the query sequence 
length			[LEN 1] Length of the alignment region in the reference sequence
positive		[LEN 2] Length of the alignment region in the query sequence (#"positive" is just a filler as blast won't all repeated fields)
pident			[% IDY] Percent identity of the alignment 
qlen			[LEN R] Length of the reference sequence 
slen			[LEN Q] Length of the query sequence 
qframe sframe	[FRM] 	Reading frame for the reference AND query sequence alignments respectively 
qseqid sseqid	[TAGS] 	The reference AND query FastA IDs respectively. All output coordinates and lengths are relative to the forward strand of the reference DNA sequence.
"""



import re
import os
from shlex import quote


def cleanID(s):
    """Remove non alphanumeric characters from string. Replace whitespace with underscores."""
    s = re.sub(r"[^\w\s]", "", s)
    s = re.sub(r"\s+", "_", s)
    return s


def _hmmbuild_command(
    exePath="hmmbuild", modelname=None, cores=None, inAlign=None, outdir=None
):
    """Construct the hmmbuild command"""
    # Make model name compliant
    modelname = cleanID(modelname)
    # Check for outdir
    if outdir:
        outdir = os.path.abspath(outdir)
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
    else:
        outdir = os.getcwd()
    # Create hmm database dir
    outdir = os.path.join(outdir, "hmmDB")
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    # Base command
    command = quote(exePath) + " --dna -n " + modelname
    # Optional set cores
    if cores:
        command += " --cpu " + str(cores)
    if outdir:
        modelout = os.path.join(outdir, modelname + ".hmm")
    else:
        modelout = modelname + ".hmm"
    # Append output file name and source alignment, return command string
    command = " ".join(
        [command, quote(os.path.abspath(modelout)), quote(os.path.abspath(inAlign))]
    )
    return command, os.path.abspath(modelout)


def _hmmpress_command(exePath="hmmpress", hmmfile=None):
    """Construct the hmmbuild command"""
    # Base command
    command = quote(exePath) + " -f " + quote(os.path.abspath(hmmfile))
    return command


def _nhmmer_command(
    exePath="nhmmer",
    modelPath=None,
    genome=None,
    evalue=None,
    nobias=False,
    matrix=None,
    cores=None,
    outdir=None,
):
    """Construct the nhmmer command"""
    # Get model hmm basename
    model_base = os.path.splitext(os.path.basename(modelPath))[0]
    # Check for outdir
    if outdir:
        outdir = os.path.abspath(outdir)
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
    else:
        outdir = os.getcwd()
    # Make subdir for nhmmer tab results
    outdir = os.path.join(outdir, "nhmmer_results")
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    # Create resultfile name
    outfile = os.path.join(os.path.abspath(outdir), model_base + ".tab")
    command = quote(exePath) + " --tblout " + quote(outfile)
    if cores:
        command += " --cpu " + str(cores)
    if evalue:
        command += " -E " + str(evalue)
    if nobias:
        command += " --nobias"
    if matrix:
        command += " --mxfile " + quote(os.path.abspath(matrix))
    command += (
        " --noali --notextw --dna --max "
        + quote(os.path.abspath(modelPath))
        + " "
        + quote(os.path.abspath(genome))
    )
    return command, outdir



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Map TIR-pHMM models to genomic sequences for annotation of MITES and
complete DNA-Transposons.
"""

import re
import os
import sys
import glob
import shutil
import tempfile
import subprocess
import pandas as pd
from Bio import SeqIO
from Bio import AlignIO
from datetime import datetime
from collections import Counter
from collections import namedtuple
from operator import attrgetter
from .hmmer_wrappers import _hmmbuild_command, _hmmpress_command, _nhmmer_command
from .bowtie2_wrappers import _bowtie2build_cmd, _bowtie2_cmd, _bam2bed_cmd
from .runBlastn import makeBlast, run_blast
from pymummer import coords_file, alignment, nucmer

from tirmite._version import __version__


class Error(Exception):
    pass


def dochecks(args):
    """Housekeeping tasks:
    Create output files/dirs and temp dirs as required."""
    # Check that specified files exist
    isfile(args.genome)
    if args.hmmFile:
        isfile(args.hmmFile)
    if args.alnFile:
        isfile(args.alnFile)
    # Make outDir if does not exist else set to current dir.
    if args.outdir:
        absOutDir = os.path.abspath(args.outdir)
        if not os.path.isdir(args.outdir):
            os.makedirs(absOutDir)
        outDir = absOutDir
    else:
        outDir = os.getcwd()
    # Make temp directory
    tempDir = os.path.join(os.getcwd(), "temp_" + getTimestring())
    os.makedirs(tempDir)
    # Return full path to output and temp directories
    return outDir, tempDir


def tSplitchecks(args):
    """Housekeeping tasks: Create output files/dirs
    and temp dirs as required."""
    if not os.path.isfile(args.infile):
        print("Input sequence file does not exist. Quitting.")
        sys.exit(1)
    # Make outDir if does not exist else set to current dir.
    if args.outdir:
        absOutDir = os.path.abspath(args.outdir)
        if not os.path.isdir(args.outdir):
            os.makedirs(absOutDir)
        outDir = absOutDir
    else:
        outDir = os.getcwd()
    # Make temp directory
    tempDir = os.path.join(os.getcwd(), "temp_" + getTimestring())
    os.makedirs(tempDir)
    # Set prefix to infile basename if none
    if not args.prefix:
        prefix = os.path.splitext(os.path.basename(args.infile))[0]
    else:
        prefix = args.prefix
    # Create outfile paths
    outfile = prefix + "_tsplit_output.fasta"
    outpath = os.path.join(outDir, outfile)
    # Return full path to output file and temp directory
    return outpath, tempDir


def importFasta2List(file):
    """Load elements from multifasta file. Check that seq IDs are unique."""
    # Read in elements from multifasta file, convert seqrecord iterator to list
    records = list(SeqIO.parse(file, "fasta"))
    # Check names are unique
    checkUniqueID(records)
    # If unique, return record list.
    return records


class Error(Exception):
    pass


def decode(x):
    try:
        s = x.decode()
    except:
        return x
    return s


def cleanID(s):
    """Remove non alphanumeric characters from string.
    Replace whitespace with underscores."""
    s = re.sub(r"[^\w\s]", "", s)
    s = re.sub(r"\s+", "_", s)
    return s


def _write_script(cmds, script):
    """Write commands into a bash script"""
    f = open(script, "w+")
    for cmd in cmds:
        print(cmd, file=f)
    f.close()


def syscall(cmd, verbose=False):
    """Manage error handling when making syscalls"""
    if verbose:
        print("Running command:", cmd, flush=True)
    try:
        output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as error:
        print(
            "The following command failed with exit code",
            error.returncode,
            file=sys.stderr,
        )
        print(cmd, file=sys.stderr)
        print("\nThe output was:\n", file=sys.stderr)
        print(error.output.decode(), file=sys.stderr)
        raise Error("Error running command:", cmd)
    if verbose:
        print(decode(output))


def run_cmd(cmds, verbose=False, tempDir=None, keeptemp=False):
    """Write and excute HMMER script"""
    if not tempDir:
        tempDir = os.getcwd()
    tmpdir = tempfile.mkdtemp(prefix="tmp.", dir=tempDir)
    original_dir = os.getcwd()
    os.chdir(tmpdir)
    script = "run_jobs.sh"
    _write_script(cmds, script)
    syscall("bash " + script, verbose=verbose)
    os.chdir(original_dir)
    if not keeptemp:
        shutil.rmtree(tmpdir)


def isfile(path):
    if not os.path.isfile(path):
        print("Input file not found: %s" % path)
        sys.exit(1)


def getTimestring():
    """Return int only string of current datetime with milliseconds."""
    (dt, micro) = datetime.utcnow().strftime("%Y%m%d%H%M%S.%f").split(".")
    dt = "%s%03d" % (dt, int(micro) / 1000)
    return dt


def checkUniqueID(records):
    """Check that IDs for input elements are unique."""
    seqIDs = [records[x].id for x in range(len(records))]
    IDcounts = Counter(seqIDs)
    duplicates = [k for k, v in IDcounts.items() if v > 1]
    if duplicates:
        print("Input sequence IDs not unique. Quiting.")
        print(duplicates)
        sys.exit(1)
    else:
        pass


def manageTemp(record=None, tempPath=None, scrub=False):
    """Create single sequence fasta files or scrub temp files."""
    if scrub and tempPath:
        try:
            os.remove(tempPath)
        except OSError:
            pass
    else:
        with open(tempPath, "w") as f:
            SeqIO.write(record, f, "fasta")


def importFasta(file):
    """Load elements from multifasta file. Check that seq IDs are unique."""
    # Read in elements from multifasta file, convert seqrecord iterator to list
    records = list(SeqIO.parse(file, "fasta"))
    # Check names are unique
    checkUniqueID(records)
    # If unique, return records as dict keyed by seq id
    recordsDict = dict()
    for rec in records:
        recordsDict[rec.id] = rec
    return recordsDict


def getbtName(file):
    """Load seqs from file. Check if more than one record.
    Report first id lable."""
    # Read in elements from fasta file, convert seqrecord iterator to list.
    records = list(SeqIO.parse(file, "fasta"))
    # Check only one seq
    if len(records) > 1:
        print(
            "Warning: %s contains multiple sequences! \n May result in \
              incorrect TIR pairing."
            % file
        )
    return cleanID(str(records[0].id))


def cmdScript(hmmDir=None, hmmFile=None, alnDir=None, tempDir=None, args=None):
    if tempDir:
        tempDir = os.path.abspath(tempDir)
        if not os.path.isdir(tempDir):
            os.makedirs(tempDir)
    else:
        tempDir = os.getcwd()
    # Create hmm database dir
    hmmDB = os.path.join(tempDir, "hmmDB")
    if not os.path.isdir(hmmDB):
        os.makedirs(hmmDB)
    # Copy all existing hmms to hmmDB
    if hmmDir:
        for hmm in glob.glob(os.path.join(hmmDir, "*.hmm")):
            shutil.copy2(hmm, hmmDB)
    if hmmFile:
        shutil.copy2(hmmFile, hmmDB)
    # Write cmds to list
    cmds = list()
    # Process alignments
    if alnDir:
        build_cmds = list()
        for alignment in glob.glob(os.path.join(alnDir, "*")):
            modelName = os.path.splitext(os.path.basename(alignment))[0]
            hmmbuildCmd, modelPath = _hmmbuild_command(
                exePath=args.hmmbuild,
                modelname=modelName,
                cores=args.cores,
                inAlign=alignment,
                outdir=tempDir,
            )
            build_cmds.append(hmmbuildCmd)
        run_cmd(build_cmds, verbose=args.verbose, keeptemp=args.keeptemp)
    # Press and write nhmmer cmd for all models in hmmDB directory
    for hmm in glob.glob(os.path.join(hmmDB, "*.hmm")):
        hmmPressCmd = _hmmpress_command(exePath=args.hmmpress, hmmfile=hmm)
        nhmmerCmd, resultDir = _nhmmer_command(
            exePath=args.nhmmer,
            nobias=args.nobias,
            matrix=args.matrix,
            modelPath=hmm,
            genome=args.genome,
            evalue=args.maxeval,
            cores=args.cores,
            outdir=tempDir,
        )
        cmds.append(hmmPressCmd)
        cmds.append(nhmmerCmd)
    # Return list of cmds and location of file result files
    return cmds, resultDir, hmmDB


def convertAlign(alnDir=None, alnFile=None, inFormat="fasta", tempDir=None):
    """Convert input alignments into Stockholm format."""
    # Construct out model path
    if tempDir:
        alnOutDir = os.path.join(os.path.abspath(tempDir), "temp_aln")
    else:
        alnOutDir = os.path.join(os.getcwd(), "temp_aln")
    # Create if does not exist
    if not os.path.isdir(alnOutDir):
        os.makedirs(alnOutDir)
    # Get list of alignment files to process
    if alnFile:
        alignments = [alnFile]
    else:
        alignments = glob.glob(alnDir)
    # Do conversion on each
    for infile in alignments:
        # Get basename
        inBase = os.path.splitext(os.path.basename(infile))[0]
        # Make outpath
        outAln = os.path.join(alnOutDir, inBase + ".stockholm")
        # Open files
        input_handle = open(infile, "r")
        output_handle = open(outAln, "w")
        # Read alignment
        alignments = AlignIO.parse(input_handle, inFormat)
        # Write as stockholm
        AlignIO.write(alignments, output_handle, "stockholm")
        # Close handles
        output_handle.close()
        input_handle.close()
    return alnOutDir


def import_nhmmer(infile=None, hitTable=None, prefix=None):
    """Read nhmmer tab files to pandas dataframe."""
    hitRecords = list()
    with open(infile, "r") as f:
        for line in f.readlines():
            li = line.strip()
            if not li.startswith("#"):
                li = li.split()
                if li[11] == "+":
                    hitRecords.append(
                        {
                            "target": li[0],
                            "model": li[2],
                            "hmmStart": li[4],
                            "hmmEnd": li[5],
                            "hitStart": li[6],
                            "hitEnd": li[7],
                            "strand": li[11],
                            "evalue": li[12],
                            "score": li[13],
                            "bias": li[14],
                        }
                    )
                elif li[11] == "-":
                    hitRecords.append(
                        {
                            "target": li[0],
                            "model": li[2],
                            "hmmStart": li[4],
                            "hmmEnd": li[5],
                            "hitStart": li[7],
                            "hitEnd": li[6],
                            "strand": li[11],
                            "evalue": li[12],
                            "score": li[13],
                            "bias": li[14],
                        }
                    )
    # Convert list of dicts into dataframe
    df = pd.DataFrame(hitRecords)
    # Reorder columns
    cols = [
        "model",
        "target",
        "hitStart",
        "hitEnd",
        "strand",
        "evalue",
        "score",
        "bias",
        "hmmStart",
        "hmmEnd",
    ]
    df = df.loc[:, cols]
    if hitTable is not None:
        # If an existing table was passed, concatenate
        df = pd.concat([df, hitTable], ignore_index=True)
    # Sort hits by HMM, Chromosome, location, and strand
    df = df.sort_values(
        ["model", "target", "hitStart", "hitEnd", "strand"],
        ascending=[True, True, True, True, True],
    )
    # Reindex
    df = df.reset_index(drop=True)
    # if prefix:
    #    df['model'] = str(prefix) + '_' + df['model'].astype(str)
    return df


def import_BED(infile=None, hitTable=None, prefix=None):
    """Read TIR bedfile to pandas dataframe."""
    # Format: Chrm, start, end, name, evalue, strand
    hitRecords = list()
    with open(infile, "r") as f:
        for line in f.readlines():
            li = line.strip()
            if not li.startswith("#"):
                li = li.split()
                hitRecords.append(
                    {
                        "target": li[0],
                        "model": li[3],
                        "hmmStart": "NA",
                        "hmmEnd": "NA",
                        "hitStart": li[1],
                        "hitEnd": li[2],
                        "strand": li[5],
                        "evalue": li[4],
                        "score": "NA",
                        "bias": "NA",
                    }
                )
    # Convert list of dicts into dataframe
    df = pd.DataFrame(hitRecords)
    # Reorder columns
    cols = [
        "model",
        "target",
        "hitStart",
        "hitEnd",
        "strand",
        "evalue",
        "score",
        "bias",
        "hmmStart",
        "hmmEnd",
    ]
    df = df.loc[:, cols]
    if hitTable is not None:
        # If an existing table was passed, concatenate
        df = pd.concat([df, hitTable], ignore_index=True)
    # Sort hits by HMM, Chromosome, location, and strand
    df = df.sort_values(
        ["model", "target", "hitStart", "hitEnd", "strand"],
        ascending=[True, True, True, True, True],
    )
    # Reindex
    df = df.reset_index(drop=True)
    # if prefix:
    #    df['model'] = str(prefix) + '_' + df['model'].astype(str)
    return df


def import_mapped(infile=None, tirName="refTIR", hitTable=None, prefix=None):
    """Read bowtie2 mapped TIR locations from bedfile to pandas dataframe."""
    print("Bowtie2 mapped reads will not be filtered on e-value.")
    hitRecords = list()
    with open(infile, "r") as f:
        for line in f.readlines():
            li = line.strip()
            if not li.startswith("#"):
                li = li.split()
                if li[3] == "+":
                    hitRecords.append(
                        {
                            "target": str(li[0]),
                            "model": tirName,
                            "hmmStart": "NA",
                            "hmmEnd": "NA",
                            "hitStart": int(li[1]),
                            "hitEnd": int(li[2]),
                            "strand": str(li[3]),
                            "evalue": 0,
                            "score": "NA",
                            "bias": "NA",
                        }
                    )
                elif li[3] == "-":
                    hitRecords.append(
                        {
                            "target": str(li[0]),
                            "model": tirName,
                            "hmmStart": "NA",
                            "hmmEnd": "NA",
                            "hitStart": int(li[1]),
                            "hitEnd": int(li[2]),
                            "strand": str(li[3]),
                            "evalue": 0,
                            "score": "NA",
                            "bias": "NA",
                        }
                    )
    # Convert list of dicts into dataframe
    df = pd.DataFrame(hitRecords)
    # Reorder columns
    cols = [
        "model",
        "target",
        "hitStart",
        "hitEnd",
        "strand",
        "evalue",
        "score",
        "bias",
        "hmmStart",
        "hmmEnd",
    ]
    df = df.loc[:, cols]
    if hitTable is not None:
        # If an existing table was passed, concatenate
        df = pd.concat([df, hitTable], ignore_index=True)
    # Sort hits by model, Chromosome, location, and strand
    df = df.sort_values(
        ["model", "target", "hitStart", "hitEnd", "strand"],
        ascending=[True, True, True, True, True],
    )
    # Reindex
    df = df.reset_index(drop=True)
    # if prefix:
    #    df['model'] = str(prefix) + '_' + df['model'].astype(str)
    return df


def filterHitsLen(hmmDB=None, mincov=None, hitTable=None):
    modelLens = dict()
    for hmm in glob.glob(os.path.join(hmmDB, "*.hmm")):
        hmmLen = None
        hmmName = None
        with open(hmm, "r") as f:
            for line in f.readlines():
                li = line.strip()
                if li.startswith("LENG"):
                    hmmLen = int(li.split()[1])
                if li.startswith("NAME"):
                    hmmName = str(li.split()[1])
            if hmmLen and hmmName:
                modelLens[hmmName] = hmmLen
    for model in modelLens.keys():
        minlen = modelLens[model] * mincov
        hitTable = hitTable.loc[
            ~(
                (hitTable["model"] == model)
                & (
                    (hitTable["hitEnd"].astype(int) - hitTable["hitStart"].astype(int))
                    + 1
                    < minlen
                )
            )
        ]
    return hitTable


def filterHitsEval(maxeval=None, hitTable=None):
    """Filter hitTable df to remove hits with e-value in excess of --maxeval."""
    hitTable = hitTable.loc[((hitTable["evalue"].astype(float)) < float(maxeval))]
    return hitTable


def table2dict(hitTable):
    """Convert pandas dataframe of nhmmer hits into dict[model][chrom]
    and index[model] = [hitlist].withCandidates and pairing status"""
    # Set up empty dict
    hitsDict = dict()
    hitIndex = dict()
    # Populate keys from dataframe
    for hmm in hitTable.model.unique():
        hitsDict[hmm] = dict()
        hitIndex[hmm] = dict()
        for chr in hitTable[hitTable["model"] == hmm].target.unique():
            hitsDict[hmm][chr] = list()
    # Set up named tuple
    hitTup = namedtuple(
        "Elem", ["model", "target", "hitStart", "hitEnd", "strand", "idx", "evalue"]
    )
    # Add each record to dicts
    for row in hitTable.iterrows():
        record = hitTup(
            row[1].model,
            row[1].target,
            int(row[1].hitStart),
            int(row[1].hitEnd),
            row[1].strand,
            row[0],
            row[1].evalue,
        )
        # Log hit for model on chromosome
        hitsDict[row[1].model][row[1].target].append(record)
        # Populate tracker
        hitIndex[hmm][row[0]] = {"rec": record, "partner": None, "candidates": list()}
    # Return master rec object and pairing tracker
    return hitsDict, hitIndex


def parseHits(hitsDict=None, hitIndex=None, maxDist=None):
    """Populate hitIndex with pairing candidates"""
    if not maxDist:
        maxDist = float("inf")
    for hmm in hitIndex.keys():
        for UID in hitIndex[hmm].keys():
            ref = hitIndex[hmm][UID]["rec"]
            if ref.strand == "+":
                for localhit in hitsDict[ref.model][ref.target]:
                    if (
                        localhit.strand == "-"
                        and localhit.hitStart >= ref.hitEnd
                        and localhit.hitStart - ref.hitEnd <= maxDist
                    ):
                        hitIndex[hmm][UID]["candidates"].append(localhit)
                # Sort candidate hit records from low to high on hitStart vals
                hitIndex[hmm][UID]["candidates"] = sorted(
                    hitIndex[hmm][UID]["candidates"],
                    key=attrgetter("hitStart", "hitEnd"),
                )
            if ref.strand == "-":
                for localhit in hitsDict[ref.model][ref.target]:
                    if (
                        localhit.strand == "+"
                        and localhit.hitEnd <= ref.hitStart
                        and ref.hitStart - localhit.hitEnd <= maxDist
                    ):
                        hitIndex[hmm][UID]["candidates"].append(localhit)
                # Sort candidate hit records from high to low on hitEnd values
                hitIndex[hmm][UID]["candidates"] = sorted(
                    hitIndex[hmm][UID]["candidates"],
                    key=attrgetter("hitEnd", "hitStart"),
                    reverse=True,
                )
    # hitIndex[model][idx].keys() == [rec,candidates,partner]
    return hitIndex


def isfirstUnpaired(ref=None, mate=None, model=None, index=None):
    """Provided with a hitID (ref) and the ID of its nearest unpaired
    candidate partner (mate), check if ref is also the nearest unpaired
    partner to mate."""
    # Init result trackers
    found = None
    mateFUP = None
    # Scan candidate partners of 'mate' looking for ref
    for matePartner in index[model][mate]["candidates"]:
        # If unpaired candidate is ref
        if not index[model][matePartner.idx]["partner"] and matePartner.idx == ref:
            found = set([matePartner.idx, mate])
            index[model][ref]["partner"] = mate
            index[model][mate]["partner"] = ref
            # return
            return found, index, mateFUP
        # If first unpaired candidate partner for mate is not ref
        elif not index[model][matePartner.idx]["partner"]:
            # Return none and unchanged index
            mateFUP = matePartner.idx
            return found, index, mateFUP
        else:
            continue
    # If mate candidates include no unpaired reps, return unchanged index
    # and blank 'found' and 'mateFUP' variables.
    return found, index, mateFUP


def getPairs(hitIndex=None, paired=None):
    """Loop over all hit for all models and search for reciprocity within
    2 degrees of the top unpaired candidate for each unpaired hit."""
    # If pair tracker not given
    if not paired:
        # Create dict of empty lists, keyed by model name
        paired = dict()
        for model in hitIndex.keys():
            paired[model] = list()
    # For each HMM model
    for model in hitIndex.keys():
        # Ask each hit in genome
        for refID in hitIndex[model].keys():
            # If it has been asigned a partner
            if not hitIndex[model][refID]["partner"]:
                # If not partnered, start checking candidate partners
                for Can1 in hitIndex[model][refID]["candidates"]:
                    # For a candidate that is also unpartnered
                    if not hitIndex[model][Can1.idx]["partner"]:
                        # Check if unpartnered candidate is a reciprocal
                        # match for our hit
                        found, hitIndex, mateFUP = isfirstUnpaired(
                            ref=refID, mate=Can1.idx, model=model, index=hitIndex
                        )
                        if found:
                            # If current hit is also the best return match of
                            # our candidate, store as pair.
                            paired[model].append(found)
                        elif mateFUP:
                            # Else if not a return match, check candidate's
                            # first outbound match for reciprocity.
                            found, hitIndex, mateFUP = isfirstUnpaired(
                                ref=Can1.idx, mate=mateFUP, model=model, index=hitIndex
                            )
                            if found:
                                # Store if found.
                                paired[model].append(found)
    return hitIndex, paired


def countUnpaired(hitIndex):
    """How many hits are still unpaired across all models."""
    count = 0
    for model in hitIndex.keys():
        for hitID in hitIndex[model].keys():
            if not hitIndex[model][hitID]["partner"]:
                count += 1
    return count


def listunpaired(hitIndex):
    """Return list of all unpaired hit IDs"""
    unpaired = list()
    for model in hitIndex.keys():
        for hitID in hitIndex[model].keys():
            if not hitIndex[model][hitID]["partner"]:
                unpaired.append(hitID)
    return unpaired


def iterateGetPairs(hitIndex, stableReps=0):
    """Iterate pairing procedure for all models until no unpaired hits remain or
    number of reps without change is exceeded."""
    # Init stable repeat counter
    reps = 0
    # Run initial pairing
    hitIndex, paired = getPairs(hitIndex=hitIndex)
    # Count remaining unpaired hits
    countUP = countUnpaired(hitIndex)
    # Iterate pairing procedure until either no unpaired remain
    # OR max number of interations without new pairing is reached
    while countUP > 0 and reps < stableReps:
        # Re-run pairing procedure
        hitIndex, paired = getPairs(hitIndex=hitIndex, paired=paired)
        # Store previous unpaired hit count
        lastCountUP = countUP
        # Update unpaired hit count
        countUP = countUnpaired(hitIndex)
        # If no change in upaired hit count, iterate stable rep counter
        if lastCountUP == countUP:
            reps += 1
    # Get IDs of remaining unpaired hits
    unpaired = listunpaired(hitIndex)
    # Return results
    return hitIndex, paired, unpaired


def extractTIRs(model=None, hitTable=None, maxeval=0.001, genome=None, padlen=None):
    """For significant hits in model, compose seqrecords."""
    # Note: Padding not yet enabled for TIR extraction.
    hitcount = 0
    seqList = list()
    for index, row in hitTable[hitTable["model"] == model].iterrows():
        if float(row["evalue"]) <= maxeval:
            hitcount += 1
            if padlen:
                hitrecord = (
                    genome[row["target"]][
                        int(row["hitStart"]) - 1 - padlen : int(row["hitStart"]) - 1
                    ].lower()
                    + genome[row["target"]][
                        int(row["hitStart"]) - 1 : int(row["hitEnd"])
                    ]
                    + genome[row["target"]][
                        int(row["hitEnd"]) : int(row["hitEnd"]) + padlen
                    ].lower()
                )
            else:
                hitrecord = genome[row["target"]][
                    int(row["hitStart"]) - 1 : int(row["hitEnd"])
                ]
            hitrecord.id = model + "_" + str(index)
            if row["strand"] == "-":
                hitrecord = hitrecord.reverse_complement(id=hitrecord.id + "_rc")
            hitrecord.name = hitrecord.id
            hitrecord.description = "_".join(
                [
                    "[" + str(row["target"]) + ":" + str(row["strand"]),
                    str(row["hitStart"]),
                    str(row["hitEnd"]) + " modelAlignment:" + row["hmmStart"],
                    row["hmmEnd"] + " E-value:" + str(row["evalue"]) + "]",
                ]
            )
            # Append record to list
            seqList.append(hitrecord)
        else:
            continue
    # Return seqrecord list and total hit count for model
    return seqList, hitcount


def writeTIRs(
    outDir=None, hitTable=None, maxeval=0.001, genome=None, prefix=None, padlen=None
):
    """Write all hits per Model to a multifasta in the outdir"""
    # Note: Padding not yet enabled for TIR extraction.
    if prefix:
        prefix = cleanID(prefix) + "_"
    else:
        prefix = ""
    if outDir:
        outDir = os.path.abspath(outDir)
        if not os.path.isdir(outDir):
            os.makedirs(outDir)
    else:
        outDir = os.getcwd()
    for model in hitTable["model"].unique():
        # List of TIR seqrecords, and count of hits
        seqList, hitcount = extractTIRs(
            model=model,
            hitTable=hitTable,
            maxeval=maxeval,
            genome=genome,
            padlen=padlen,
        )
        outfile = os.path.join(
            outDir, prefix + model + "_hits_" + str(hitcount) + ".fasta"
        )
        # Write extracted hits to model outfile
        with open(outfile, "w") as handle:
            for seq in seqList:
                seq.id = prefix + str(seq.id)
                SeqIO.write(seq, handle, "fasta")


# CS10_Chromosome_02_+_88294_88353_modelAlignment:1_60


def flipTIRs(x, y):
    """Sort hits into left and right TIRs."""
    left2right = sorted([x, y], key=attrgetter("hitStart", "hitEnd"))
    return (left2right[0], left2right[1])


def fetchElements(paired=None, hitIndex=None, genome=None):
    """Extract complete sequence of paired elements,
    asign names and child TIRs for use in seq and GFF reporting."""
    TIRelements = dict()
    gffTup = namedtuple(
        "gffElem",
        [
            "model",
            "chromosome",
            "start",
            "end",
            "strand",
            "type",
            "id",
            "leftHit",
            "rightHit",
            "seq",
            "evalue",
        ],
    )
    for model in paired.keys():
        TIRelements[model] = list()
        model_counter = 0
        for x, y in paired[model]:
            model_counter += 1
            x = hitIndex[model][x]["rec"]
            y = hitIndex[model][y]["rec"]
            leftHit, rightHit = flipTIRs(x, y)
            eleID = model + "_Element_" + str(model_counter)
            eleSeq = genome[leftHit.target][
                int(leftHit.hitStart) - 1 : int(rightHit.hitEnd)
            ]
            eleSeq.id = eleID
            eleSeq.name = eleID
            eleSeq.description = (
                "_".join(
                    [
                        "[" + leftHit.target + ":" + str(leftHit.hitStart),
                        str(rightHit.hitEnd),
                    ]
                )
                + " len="
                + str(rightHit.hitEnd - leftHit.hitStart)
                + "]"
            )

            TIRelement = gffTup(
                model,
                leftHit.target,
                leftHit.hitStart,
                rightHit.hitEnd,
                leftHit.strand,
                "TIR_Element",
                eleID,
                leftHit,
                rightHit,
                eleSeq,
                "NA",
            )

            TIRelements[model].append(TIRelement)
    # Return list of element info tuples
    return TIRelements


def writeElements(outDir, eleDict=None, prefix=None):
    """Takes dict of extracted sequences keyed by model.
    Writes to fasta by model."""
    if prefix:
        prefix = cleanID(prefix) + "_"
    else:
        prefix = ""
    for model in eleDict.keys():
        outfile = os.path.join(outDir, prefix + model + "_elements.fasta")
        with open(outfile, "w") as handle:
            for element in eleDict[model]:
                element.seq.id = prefix + str(element.seq.id)
                SeqIO.write(element.seq, handle, "fasta")


def writePairedTIRs(
    outDir=None, paired=None, hitIndex=None, genome=None, prefix=None, padlen=None
):
    """Extract TIR sequence of paired hits, write to fasta."""
    # Note: Sequence padding not yet enabled for paired TIRs
    TIRpairs = dict()
    gffTup = namedtuple(
        "gffElem",
        [
            "model",
            "chromosome",
            "start",
            "end",
            "strand",
            "type",
            "id",
            "leftHit",
            "rightHit",
            "seq",
            "evalue",
        ],
    )
    for model in paired.keys():
        TIRpairs[model] = list()
        model_counter = 0
        for x, y in paired[model]:
            model_counter += 1
            x = hitIndex[model][x]["rec"]
            y = hitIndex[model][y]["rec"]
            leftHit, rightHit = flipTIRs(x, y)
            eleID = model + "_TIRpair_" + str(model_counter)
            # If padlen set, extract hit with x bases either side
            if padlen:
                eleSeqLeft = (
                    genome[leftHit.target][
                        int(leftHit.hitStart) - 1 - padlen : int(leftHit.hitStart) - 1
                    ].lower()
                    + genome[leftHit.target][
                        int(leftHit.hitStart) - 1 : int(leftHit.hitEnd)
                    ]
                    + genome[leftHit.target][
                        int(leftHit.hitEnd) : int(leftHit.hitEnd) + padlen
                    ].lower()
                )
                eleSeqRight = (
                    genome[leftHit.target][
                        int(rightHit.hitStart) - 1 - padlen : int(rightHit.hitStart) - 1
                    ].lower()
                    + genome[leftHit.target][
                        int(rightHit.hitStart) - 1 : int(rightHit.hitEnd)
                    ]
                    + genome[leftHit.target][
                        int(rightHit.hitEnd) : int(rightHit.hitEnd) + padlen
                    ].lower()
                )
            else:
                eleSeqLeft = genome[leftHit.target][
                    int(leftHit.hitStart) - 1 : int(leftHit.hitEnd)
                ]
                eleSeqRight = genome[leftHit.target][
                    int(rightHit.hitStart) - 1 : int(rightHit.hitEnd)
                ]
            eleSeqRight = eleSeqRight.reverse_complement()
            eleSeqLeft.id = eleID + "_L"
            eleSeqLeft.name = eleID + "_L"
            eleSeqLeft.description = (
                "_".join(
                    [
                        "[" + leftHit.target + ":" + str(leftHit.hitStart),
                        str(leftHit.hitEnd),
                    ]
                )
                + "]"
            )
            eleSeqRight.id = eleID + "_R"
            eleSeqRight.name = eleID + "_R"
            eleSeqRight.description = (
                "_".join(
                    [
                        "[" + leftHit.target + ":" + str(rightHit.hitEnd),
                        str(rightHit.hitStart),
                    ]
                )
                + "]"
            )
            TIRleft = gffTup(
                model,
                leftHit.target,
                leftHit.hitStart,
                leftHit.hitEnd,
                leftHit.strand,
                "TIR",
                eleSeqLeft.id,
                leftHit,
                rightHit,
                eleSeqLeft,
                "NA",
            )
            TIRright = gffTup(
                model,
                leftHit.target,
                rightHit.hitStart,
                rightHit.hitEnd,
                leftHit.strand,
                "TIR",
                eleSeqRight.id,
                leftHit,
                rightHit,
                eleSeqRight,
                "NA",
            )
            TIRpairs[model].append(TIRleft)
            TIRpairs[model].append(TIRright)
    if prefix:
        prefix = cleanID(prefix) + "_"
    else:
        prefix = ""
    for model in TIRpairs.keys():
        outfile = os.path.join(
            outDir,
            prefix + model + "_paired_TIR_hits_" + str(model_counter * 2) + ".fasta",
        )
        with open(outfile, "w") as handle:
            for element in TIRpairs[model]:
                element.seq.id = prefix + str(element.seq.id)
                SeqIO.write(element.seq, handle, "fasta")


def fetchUnpaired(hitIndex=None):
    """Take list of unpaired hit IDs from listunpaired(),
    Compose TIR gff3 record."""
    orphans = list()
    gffTup = namedtuple(
        "gffElem",
        [
            "model",
            "chromosome",
            "start",
            "end",
            "strand",
            "type",
            "id",
            "leftHit",
            "rightHit",
            "seq",
            "evalue",
        ],
    )
    for model in hitIndex.keys():
        for recID in hitIndex[model].keys():
            if not hitIndex[model][recID]["partner"]:
                x = hitIndex[model][recID]["rec"]
                orphan = gffTup(
                    x.model,
                    x.target,
                    x.hitStart,
                    x.hitEnd,
                    x.strand,
                    "orphan_TIR",
                    x.idx,
                    None,
                    None,
                    None,
                    x.evalue,
                )
                orphans.append(orphan)
    return orphans


def gffWrite(
    outpath=None,
    featureList=list(),
    writeTIRs=True,
    unpaired=None,
    suppressMeta=False,
    prefix=None,
):
    """
    Write predicted paired-TIR features (i.e. MITEs) from fetchElements()
    as GFF3. Optionally, write child TIRS and orphan TIRs to GFF3 also.
    """
    if prefix:
        prefix = cleanID(prefix) + "_"
    else:
        prefix = ""
    # If path to output gff3 file not provided, set default location.
    if not outpath:
        outpath = os.path.join(os.getcwd(), "tirmite_features.gff3")
    # Unpack element dict to list
    allfeatures = list()
    for model in featureList.keys():
        for record in featureList[model]:
            allfeatures.append(record)
    # Add list of unpaired TIRs to main featureList if provided.
    if unpaired:
        allfeatures = allfeatures + unpaired
    # Sort features
    sortedFeatures = sorted(
        allfeatures, key=attrgetter("model", "chromosome", "start", "end")
    )
    # Open GFF handle
    with open(outpath, "w") as file:
        # Write headers
        file.write("##gff-version 3" + "\n")
        file.write(
            "\t".join(
                [
                    "#seqid",
                    "source",
                    "type",
                    "start",
                    "end",
                    "score",
                    "strand",
                    "phase",
                    "attributes",
                ]
            )
            + "\n"
        )
        # Format features for GFF3
        for x in sortedFeatures:
            if x.type == "orphan_TIR" and writeTIRs in ["all", "unpaired"]:
                file.write(
                    "\t".join(
                        [
                            str(x.chromosome),
                            "tirmite",
                            x.type,
                            str(x.start),
                            str(x.end),
                            ".",
                            x.strand,
                            ".",
                            "ID="
                            + prefix
                            + str(x.model)
                            + "_"
                            + str(x.id)
                            + ";model="
                            + str(x.model)
                            + ";evalue="
                            + str(x.evalue)
                            + ";",
                        ]
                    )
                    + "\n"
                )
            if x.type == "TIR_Element":
                # Write Element line
                file.write(
                    "\t".join(
                        [
                            str(x.chromosome),
                            "tirmite",
                            x.type,
                            str(x.start),
                            str(x.end),
                            ".",
                            x.strand,
                            ".",
                            "ID=" + prefix + str(x.id) + ";model=" + str(x.model) + ";",
                        ]
                    )
                    + "\n"
                )
                if writeTIRs in ["all", "paired"]:
                    # Write left TIR line as child
                    l = x.leftHit
                    file.write(
                        "\t".join(
                            [
                                str(l.target),
                                "tirmite",
                                "paired_TIR",
                                str(l.hitStart),
                                str(l.hitEnd),
                                ".",
                                l.strand,
                                ".",
                                "ID="
                                + prefix
                                + str(x.model)
                                + "_"
                                + str(l.idx)
                                + ";model="
                                + str(x.model)
                                + ";Parent="
                                + str(x.id)
                                + ";evalue="
                                + str(l.evalue)
                                + ";",
                            ]
                        )
                        + "\n"
                    )
                    # Write right TIR line as child on neg strand
                    r = x.rightHit
                    file.write(
                        "\t".join(
                            [
                                str(r.target),
                                "tirmite",
                                "paired_TIR",
                                str(r.hitStart),
                                str(r.hitEnd),
                                ".",
                                r.strand,
                                ".",
                                "ID="
                                + prefix
                                + str(x.model)
                                + "_"
                                + str(r.idx)
                                + ";model="
                                + str(x.model)
                                + ";Parent="
                                + str(x.id)
                                + ";evalue="
                                + str(r.evalue)
                                + ";",
                            ]
                        )
                        + "\n"
                    )


def getLTRs(
    elements=None,
    flankdist=10,
    minid=80,
    minterm=10,
    minseed=5,
    diagfactor=0.3,
    report="split",
    temp=None,
    keeptemp=False,
    alignTool="nucmer",
    verbose=False,
):
    """Align elements to self and attempt to identify LTRs."""
    # Set temp directory to cwd if none.
    if not temp:
        temp = os.getcwd()
    # For each candidate LTR element
    for rec in elements:
        # Create temp paths for single element fasta and alignment coords
        tempFasta = os.path.join(temp, cleanID(rec.id) + ".fasta")
        tempCoords = tempFasta + ".coords"
        # Write current element to single fasta
        manageTemp(record=rec, tempPath=tempFasta, scrub=False)
        # Align to self with nucmer
        if alignTool == "nucmer":
            # Compose Nucmer script for current element vs self
            runner = nucmer.Runner(
                tempFasta,
                tempFasta,
                tempCoords,
                min_id=minid,
                min_length=minseed,
                diagfactor=diagfactor,
                mincluster=minterm,
                breaklen=200,
                maxmatch=True,
                simplify=False,
            )
            # Execute nucmer
            runner.run()
        elif alignTool == "blastn":
            # Alternatively, use blastn as search tool and write
            # nucmer.coords-like output.
            cmds = makeBlast(seq=tempFasta, outfile=tempCoords, pid=minid)
            run_blast(cmds, verbose=verbose)
        # Import coords file to iterator object
        file_reader = coords_file.reader(tempCoords)
        # Exclude hits to self. Also converts iterator output to stable list
        alignments = [hit for hit in file_reader if not hit.is_self_hit()]
        # Filter hits less than min length (Done internally for nucmer,
        # not blastn.)
        alignments = [
            hit for hit in alignments if hit.ref_end - hit.ref_start >= minterm
        ]
        # Filter for hits on same strand i.e. tandem repeats / LTRs
        alignments = [hit for hit in alignments if hit.on_same_strand()]
        # Filter for 5' repeats which begin within x bases of element start
        alignments = [hit for hit in alignments if hit.ref_start <= flankdist]
        # Filter for 5' repeats whose 3' match ends within x bases of element end
        alignments = [hit for hit in alignments if len(rec) - hit.qry_end <= flankdist]
        # Scrub overlappying ref / query segments, and also complementary
        # 3' to 5' flank hits
        alignments = [hit for hit in alignments if hit.ref_end < hit.qry_start]
        # Sort largest to smallest dist between end of ref (subject) and start
        # of query (hit)
        # x.qry_start (3') - x.ref_end (5') = Length of internal segment
        # Note: Need to check that this sorts correctlly for alignments is same
        # orientation.
        print("Warning: Check candidate pairs are correctly sorted.")
        alignments = sorted(
            alignments, key=lambda x: (x.qry_start - x.ref_end), reverse=True
        )
        # If alignments exist after filtering report features using alignment
        # pair with largest internal segment i.e. first element in sorted list.
        if alignments:
            if verbose:
                [print(x) for x in alignments]
            if report == "all":
                # yield original element
                yield rec
            if report in ["split", "external"]:
                # yield LTR slice - append "_LTR"
                extSeg = rec[alignments[0].ref_start : alignments[0].ref_end + 1]
                extSeg.id = extSeg.id + "_LTR"
                extSeg.name = extSeg.id
                extSeg.description = "[" + rec.id + " LTR segment]"
                yield extSeg
            if report in ["split", "internal"]:
                # yield internal slice - append "_I"
                intSeg = rec[alignments[0].ref_end : alignments[0].qry_start + 1]
                intSeg.id = intSeg.id + "_I"
                intSeg.name = intSeg.id
                intSeg.description = "[" + rec.id + " internal segment]"
                yield intSeg
        else:
            # If alignment list empty after filtering print alert and continue
            print("No LTRs found for candidate element: %s" % rec.id)
        # Scrub single fasta and coords file for current element.
        if not keeptemp:
            manageTemp(tempPath=tempFasta, scrub=True)
            manageTemp(tempPath=tempCoords, scrub=True)


def getTIRs(
    elements=None,
    flankdist=2,
    minid=80,
    minterm=10,
    minseed=5,
    diagfactor=0.3,
    mites=False,
    report="split",
    temp=None,
    keeptemp=False,
    alignTool="nucmer",
    verbose=False,
):
    """
    Align elements to self and attempt to identify TIRs.
    Optionally attempt to construct synthetic MITEs from TIRs.
    """
    # Set temp directory to cwd if none.
    if not temp:
        temp = os.getcwd()
    # For each candidate LTR element
    for rec in elements:
        # Create temp paths for single element fasta and alignment coords
        tempFasta = os.path.join(temp, cleanID(rec.id) + ".fasta")
        tempCoords = tempFasta + ".coords"
        # Write current element to single fasta
        manageTemp(record=rec, tempPath=tempFasta, scrub=False)
        # Align to self with nucmer
        if alignTool == "nucmer":
            # Compose Nucmer script for current element vs self
            runner = nucmer.Runner(
                tempFasta,
                tempFasta,
                tempCoords,
                min_id=minid,
                min_length=minseed,
                diagfactor=diagfactor,
                mincluster=minterm,
                breaklen=200,
                maxmatch=True,
                simplify=False,
            )
            # Execute nucmer
            runner.run()
        elif alignTool == "blastn":
            # Alternatively, use blastn as search tool and write nucmer.coords-like output.
            cmds = makeBlast(seq=tempFasta, outfile=tempCoords, pid=minid)
            run_blast(cmds, verbose=verbose)
        # Import coords file to iterator object
        file_reader = coords_file.reader(tempCoords)
        # Exclude hits to self. Also converts iterator output to stable list
        alignments = [hit for hit in file_reader if not hit.is_self_hit()]
        # Filter hits less than min length (Done internally for nucmer, not blastn.)
        alignments = [
            hit for hit in alignments if hit.ref_end - hit.ref_start >= minterm
        ]
        # Filter for hits on same strand i.e. tandem repeats / LTRs
        alignments = [hit for hit in alignments if not hit.on_same_strand()]
        # Filter for 5' repeats which begin within x bases of element start
        alignments = [hit for hit in alignments if hit.ref_start <= flankdist]
        # Scrub overlappying ref / query segments, and also complementary
        # 3' to 5' flank hits
        alignments = [hit for hit in alignments if hit.ref_end < hit.qry_end]
        # Sort largest to smallest dist between end of ref (subject) and start
        # of query (hit)
        # x.qry_end - x.ref_end = 5'end of right TIR - 3' end of left
        # TIR = length of internal segment
        # TIR pair with smallest internal segment (longest TIRs) is first in list.
        alignments = sorted(
            alignments, key=lambda x: (x.qry_end - x.ref_end), reverse=False
        )
        # If alignments exist after filtering report features using alignment
        # pair with largest internal segment i.e. first element in sorted list.
        if alignments:
            if verbose:
                [print(x) for x in alignments]
            if report in ["split", "external", "all"]:
                # yield TIR slice - append "_TIR"
                extSeg = rec[alignments[0].ref_start : alignments[0].ref_end + 1]
                extSeg.id = extSeg.id + "_TIR"
                extSeg.name = extSeg.id
                extSeg.description = "[" + rec.id + " TIR segment]"
                yield extSeg
            if report in ["split", "internal", "all"]:
                # yield internal slice - append "_I"
                intSeg = rec[alignments[0].ref_end : alignments[0].qry_end + 1]
                intSeg.id = intSeg.id + "_I"
                intSeg.name = intSeg.id
                intSeg.description = "[" + rec.id + " internal segment]"
                yield intSeg
            if report == "all":
                yield rec
            if mites:
                # Assemble TIRs into hypothetical MITEs
                synMITE = (
                    rec[alignments[0].ref_start : alignments[0].ref_end + 1]
                    + rec[alignments[0].qry_end : alignments[0].qry_start + 1]
                )
                synMITE.id = synMITE.id + "_synMITE"
                synMITE.name = synMITE.id
                synMITE.description = (
                    "[Synthetic MITE constructed from " + rec.id + " TIRs]"
                )
                yield synMITE
        else:
            # If alignment list empty after filtering print alert and continue
            print("No TIRs found for candidate element: %s" % rec.id)
        # Scrub single fasta and coords file for current element.
        if not keeptemp:
            manageTemp(tempPath=tempFasta, scrub=True)
            manageTemp(tempPath=tempCoords, scrub=True)


def segWrite(outfile, segs=None):
    """
    Take a generator object yielding seqrecords and
    write each to outfile in fasta format.
    """
    seqcount = 0
    if segs:
        with open(outfile, "w") as handle:
            for seq in segs:
                seqcount += 1
                SeqIO.write(seq, handle, "fasta")
        if seqcount == 0:
            os.remove(outfile)


# gffTup fields: 'model', 'chromosome', 'start', 'end', 'strand', 'type', 'id', 'score','bias', 'evalue', 'leftHit' , 'rightHit', 'eleSeq'
# Types: "TIR_Element", "orphan_TIR"


"""
# Useful attributes of pymummer objects:
[x.ref_start for x in alignments]
[x.ref_end for x in alignments]
[x.qry_start for x in alignments]
[x.qry_end for x in alignments]
[x.hit_length_ref for x in alignments]
[x.hit_length_qry for x in alignments]
[x.percent_identity for x in alignments]
[x.ref_length for x in alignments]
[x.qry_length for x in alignments]
[x.frame for x in alignments]
[x.ref_name for x in alignments]
[x.qry_name for x in alignments]

## Don't use these, bizzaro format. Not indexed to 0. Cannot sort as ints.
#coord.reverse_query()
#coord.reverse_reference()
#coord.qry_coords()
#coord.ref_coords()
"""
